package arraybasedDS;

/**
 * @author Luca Castelli Aleardi
 * Array based implementation of a halfedge representation of a triangle mesh
 * References are encoded with integer values (32 bits), stores in an array of int
 */
public class ArrayBasedHalfedge implements HalfedgeInterface, NavigationInterface {
	
	private int nVertices, nHalfedges; // number of vertices and halfedges in the mesh
	private int[] next, opposite, vertex; // each halfedge is represented with three integer references
	private int[] incidentEdge; // for each vertex we store an incident (ingoing) halfedge

	private float[][] coordF; // geometric coordinates (for testing performances without java interfaces)
	
	public ArrayBasedHalfedge(int nVertices, int nHalfedges) {
		this.nVertices=nVertices;
		this.nHalfedges=nHalfedges;
		this.next=new int[nHalfedges];
		this.opposite=new int[nHalfedges];
		this.vertex=new int[nHalfedges];
		this.incidentEdge=new int[nVertices];

		// geometric coordinates
		this.coordF=new float[nVertices][3]; // coordinates simple float precision
	}
	
	/** 
	 * Return the opposite of an half edge 
	 **/
	public void setOpposite(int e, int eOpposite) { 
		opposite[e]=eOpposite;
	}
	
	public void setNext(int e, int eNext) {
		next[e]=eNext;
	}
	
	public void setVertex(int e, int v) {
		vertex[e]=v;
	}
	
	public void setEdge(int v, int e) {
		incidentEdge[v]=e;
	}

	public void setPoint(int v, float x, float y, float z) {
		this.coordF[v][0]=x;
		this.coordF[v][1]=y;
		this.coordF[v][2]=z;			
	}
	
	//--- methods for accessing data and navigating between mesh elements ---

	/** 
	 * Return the following edge in ccw orientation around the same face
	 **/
	public int getNext(int e) {
		return next[e];
	}
	
	/** 
	 * Return the opposite edge in the neighboring face (having opposite direction)
	 **/
	public int getOpposite(int e) {
		return opposite[e];
	}
	
	/** 
	 * Return the previous edge in ccw orientation around the same face
	 **/
	public int getPrev(int e) {
		return next[next[e]];
	}
	
	/** 
	 * Return the target vertex incident to the half edge
	 **/
	public int getTarget(int e) {
		return vertex[e];
	}
	
	/** 
	 * Return the origin vertex incident to the half edge
	 **/
	public int getSource(int e) {
		return vertex[next[e]];
	}
	
	/** 
	 * Return a half edge incident to the vertex
	 **/
	public int getEdge(int v) {
		return incidentEdge[v];
	}
	
	/** 
	 * Return geometric position of a vertex
	 **/
	public float[] getPoint(int v) {
		return this.coordF[v];
	}
	
	/** 
	 * Return the number of vertices
	 **/
	public int sizeOfVertices() {
		return this.nVertices;
	}
	
	/** 
	 * Return the number of half edges
	 **/
	public int sizeOfHalfedges() {
		return this.nHalfedges;
	}
	
	/** 
	 * Return the list of (geometric points) neighbors of a given vertex
	 **/
	public int[] getNeighborsArray(int v) {
		int[] result=new int[45];
		int cont=0;
		
		int edge=this.getEdge(v);
		int pEdge=this.getNext(this.getOpposite(edge));
		while(pEdge!=edge) {
			result[cont]=this.getSource(pEdge);
			pEdge=this.getNext(this.getOpposite(pEdge));
			cont++;
		}
		result[cont]=this.getSource(pEdge);
		return result;
	}

	/** 
	 * Return the degree of a given vertex
	 **/
	public int vertexDegree(int v) {
		int e=this.incidentEdge[v];
		int result=0;
		
		int pEdge=this.opposite[next[e]];
		while(pEdge!=e) {
			pEdge=this.opposite[next[pEdge]];
			result++;
		}
		return result+1;
	}
	
	/** 
	 * Return the degree of a given vertex
	 * Slow version: with function calls
	 **/
	public int vertexDegreeSlow(int v) {
		int e=this.getEdge(v);
		int result=0;
		
		int pEdge=this.getOpposite(this.getNext(e));
		while(pEdge!=e) {
			pEdge=this.getOpposite(this.getNext(pEdge));
			result++;
		}
		return result+1;
	}
	
	/**
	 * Return the normal to a vertex in the mesh
	 * Fast version with simple Float precision, with no use of Point_ interface
	 */
	public float[] getNormalFloat(int v) {
		throw new Error("To be completed");
		/*
		float[] result=new float[3]; // average of cross product vectors
		float[] prod; // cross product vector
		int degree=1; // vertex degree
		float[] pV, pU, pW; // coordinates of vertices v, u and w
		float[] uv=new float[3]; // coordinates of vector V-U
		float[] uw=new float[3]; // coordinates of vector W-U

		int e=this.incidentEdge[v]; // half edge (u,v) oriented toward v
		int u=this.vertex[opposite[e]]; // vertex u is the source of edge e
		int pEdge=this.opposite[this.next[e]]; // half edge (v,w) turning around vertex v
		int w=this.vertex[this.opposite[pEdge]];

		pV=this.coordF[v]; // coordinates of vertex v
		pU=this.coordF[u]; // coordinates of vertex u
		pW=this.coordF[w]; // coordinates of vertex w		

		//System.out.println(vectorToString(pV)+" - "+vectorToString(pU)+" - "+vectorToString(pW)); // debug

		// computing vectors V-U and W-U
		uv[0]=pV[0]-pU[0]; uv[1]=pV[1]-pU[1]; uv[2]=pV[2]-pU[2];
		uw[0]=pW[0]-pU[0]; uw[1]=pW[1]-pU[1]; uw[2]=pW[2]-pU[2];	
		result=NavigationInterface.crossProduct(uv, uw);
		pU=pW;
		uv=uw;
		
		
		while(pEdge!=e) { // turning around vertex v
			pEdge=this.opposite[this.next[pEdge]];
			w=this.vertex[this.opposite[pEdge]];
			pW=this.coordF[w]; // coordinates of vertex w
			
			//System.out.println(vectorToString(pV)+" - "+vectorToString(pU)+" - "+vectorToString(pW)); // debug

			uw[0]=pW[0]-pU[0]; 
			uw[1]=pW[1]-pU[1]; 
			uw[2]=pW[2]-pU[2];
			prod=NavigationInterface.crossProduct(uv, uw); // computing cross product of the 2 vectors
			
			result[0]=result[0]+prod[0];
			result[1]=result[1]+prod[1];
			result[2]=result[2]+prod[2];
			
			pU=pW;
			uv=uw;
			degree++;
		}
		
		result[0]=result[0]/degree;
		result[1]=result[1]/degree;
		result[2]=result[2]/degree;
		
		//System.out.println("degree: "+degree);
		return result;
		*/
	}
	
	  public String vectorToString(float[] u) {
		  return "("+u[0]+","+u[1]+","+u[2]+")";
	  }

		/**
		 * Return the name of the data structure
		 */
		public String name() {
			return "Half edge";
		}

}
