package arraybasedDS;

import Jcg.geometry.*;
import Jcg.polyhedron.*;

/**
 * @author Luca Castelli Aleardi
 * This class provides functions for creating array based implementations of mesh data structures
 * (from input data or from existing mesh data structures)
 */
public class ArrayBasedMeshBuilder<X extends Point_> {
	
	Polyhedron_3<X> polyhedron;
	
	public ArrayBasedMeshBuilder(Polyhedron_3<X> polyhedron) {
		this.polyhedron=polyhedron;		
	}

	/**
	 * Construct an array based implementation of a halfedge data structure.
	 * Recall that Polyhedron_3 is a (pointer based) halfedge DS for polygonal meshes
	 */
	public ArrayBasedHalfedge halfedgeFromPolyhedron(){
		System.out.print("Creating an array based half edge DS from a Polyhedron...");
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances

    	int nV=this.polyhedron.sizeOfVertices();
		int nE=this.polyhedron.sizeOfHalfedges();
		ArrayBasedHalfedge result=new ArrayBasedHalfedge(nV, nE);
		
		// setting vertex and edge labels
		int i=0;
		for(Vertex<X> v: this.polyhedron.vertices) {
			v.index=i;
			i++;
		}
		i=0;
		for(Halfedge<X> e:this.polyhedron.halfedges) {
			e.index=i;
			i++;
		}
		
		// creating the ArrayBased implementation of the halfedge DS
		for(Halfedge<X> e: this.polyhedron.halfedges) {
			int edgeLabel=e.index;

			int nextLabel=-1;
			int oppositeLabel=-1;
			int originLabel=-1;
			
			if(e.getNext()!=null)
				nextLabel=e.getNext().index;
			
			if(e.getOpposite()!=null)
				oppositeLabel=e.getOpposite().index;
			
			if(e.getVertex()!=null)
				originLabel=e.getVertex().index;
			
			result.setNext(edgeLabel, nextLabel);
			result.setOpposite(edgeLabel, oppositeLabel);
			result.setVertex(edgeLabel, originLabel);			
		}
		
		//System.out.println("initializing vertices and points");
		for(Vertex<X> v: this.polyhedron.vertices) {
			result.setEdge(v.index, v.getHalfedge().index);
			float x=polyhedron.getX(v.index);
			float y=polyhedron.getY(v.index);
			float z=polyhedron.getZ(v.index);
			result.setPoint(v.index, x, y, z);
		}
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
		System.out.println("done ("+duration+" s)");
		
		return result;
	}

	/**
	 * Construct an array based implementation of a winged edge data structure.
	 * from an oriented graph (encoded as half-edge data structure)
	 * Half-edges must be oriented. 
	 * Half-edges must have labels between 0..2E-1
	 * Vertices must have labels between 0..N-1
	 */
	public ArrayBasedWingededge wingededgeFromOrientedPolyhedron(boolean[] orientation){
		System.out.print("Creating an array based winged edge DS from a Polyhedron...");
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances

    	int nV=this.polyhedron.sizeOfVertices();
		int nE=this.polyhedron.sizeOfHalfedges()/2;
		ArrayBasedWingededge result=new ArrayBasedWingededge(nV, nE);
		
		// conversion from half-edge labels to edge labels
		// every pair of opposite half-edges must have the same label
		int[] edgeIndex=new int[this.polyhedron.sizeOfHalfedges()];
		
		// checking vertex labels
		int i=0;
		for(Vertex<X> v: this.polyhedron.vertices) {
			if(v.index<0 || v.index>=polyhedron.sizeOfVertices())
				throw new Error("wrong vertex labels: "+v.index);
			i++;
		}

		// checking half-edge labels and edge orientation
		i=0;
		for(Halfedge<X> e:this.polyhedron.halfedges) {
			if(e.index<0)
				throw new Error("wrong edge labels: "+e.index);
			if(orientation[e.index]==orientation[e.getOpposite().index])
				throw new Error("wrong half-edge orientation: "+e.index);
			i++;
		}

		// setting new edge labels from half-edges labels
		i=0;
		for(Halfedge<X> e:this.polyhedron.halfedges) {
			if(orientation[e.index]==true) {
				edgeIndex[e.index]=i;
				i++;
			}
		}

		i=0;
		for(Halfedge<X> e:this.polyhedron.halfedges) {
			if(orientation[e.index]==false) {
				edgeIndex[e.index]=edgeIndex[e.getOpposite().index];
				i++;
			}
		}
		
		// creating the Array Based implementation of the winged edge DS
		//System.out.println("Initializing edges");
		
		// initializing edges
		for(Halfedge<X> e: this.polyhedron.halfedges) {
			int eLabel=edgeIndex[e.index];
			if(orientation[e.index]==true) {
				result.setLeftFront(eLabel, edgeIndex[e.getNext().index]);
				result.setLeftBack(eLabel, edgeIndex[e.getPrev().index]);
				result.setRightFront(eLabel, edgeIndex[e.getOpposite().getPrev().index]);
				result.setRightBack(eLabel, edgeIndex[e.getOpposite().getNext().index]);
				result.setSource(eLabel, e.getOpposite().getVertex().index);
				result.setTarget(eLabel, e.getVertex().index);
			}
		}
		
		// initializing vertices and points
		for(Vertex<X> v: this.polyhedron.vertices) {
			result.setEdge(v.index, edgeIndex[v.getHalfedge().index]);
			float x=polyhedron.getX(v.index);
			float y=polyhedron.getY(v.index);
			float z=polyhedron.getZ(v.index);
			result.setPoint(v.index, x, y, z);
		}
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
		System.out.println("done ("+duration+" s)");
		
		return result;
	}
	
	/**
	 * Check whether a table of n integers encode a permutation of order n
	 * Every integer between 0 and N-1 must appear exactly once in table t
	 */
	public static boolean checkPermutation(int[] t) {
		byte[] occurrences=new byte[t.length];
		for(int i=0;i<t.length;i++) {
			if(t[i]<0 || t[i]>=t.length)
				return false;
			occurrences[t[i]]++;
		}
		for(int i=0;i<occurrences.length;i++)
			if(occurrences[i]!=1)
				return false;
		return true;
	}

}
