package arraybasedDS;

/**
 * @author Luca Castelli Aleardi
 * Array based implementation of the winged edge data structure
 * References are encoded with integer values (32 bits)
 */
public class ArrayBasedWingededge implements WingededgeInterface, NavigationInterface {

	private int nVertices, nEdges; // number of vertices and edges in the mesh
	private int[] leftFront, rightFront, leftBack, rightBack; // each edge is represented with four references
	private int[] source, target; // every edge stores two references to the incident vertices
	private int[] incidentEdge; // for each vertex we store an incident (outgoing) halfedge

	private float[][] coordF; // geometric coordinates

	public ArrayBasedWingededge(int nVertices, int nEdges) {
		this.nVertices=nVertices;
		this.nEdges=nEdges;
		this.leftFront=new int[nEdges];
		this.rightFront=new int[nEdges];
		this.leftBack=new int[nEdges];
		this.rightBack=new int[nEdges];
		
		this.source=new int[nEdges];
		this.target=new int[nEdges];
		this.incidentEdge=new int[nVertices];
		
		this.coordF=new float[nVertices][3]; // setting geometric coordinates
	}
	
	public void setLeftFront(int e, int eLeftFront) { leftFront[e]=eLeftFront; }
	public void setRightFront(int e, int eRightFront) { rightFront[e]=eRightFront; }
	public void setLeftBack(int e, int eLeftBack) { leftBack[e]=eLeftBack; }
	public void setRightBack(int e, int eRightBack) { rightBack[e]=eRightBack; }

	public void setSource(int e, int v) { source[e]=v; }
	public void setTarget(int e, int v) { target[e]=v; }
	
	public void setEdge(int v, int e) { incidentEdge[v]=e; }
	
	public void setPoint(int v, float x, float y, float z) {
		this.coordF[v][0]=x;
		this.coordF[v][1]=y;
		this.coordF[v][2]=z;			
	}

	//--- methods for accessing data and navigation in the mesh ---
	
	public int sizeOfVertices() { return this.nVertices; }
	public int sizeOfEdges() { return this.nEdges; }
	
	public int getLeftFront(int e) { return leftFront[e]; }
	public int getRightFront(int e) { return rightFront[e]; }
	public int getLeftBack(int e) { return leftBack[e]; }
	public int getRightBack(int e) { return rightBack[e]; }
	
	public int getSource(int e) { return source[e]; }
	public int getTarget(int e) { return target[e];}
	
	public int getEdge(int v) { return incidentEdge[v]; }	
	public float[] getPoint(int v) { return this.coordF[v]; }
		
	/** 
	 * Check whether edge e is outgoing from vertex v
	 **/
	public boolean isOutgoing(int e, int v) {
		if(this.getSource(e)==v) return true;
		return false;
	}

	/** 
	 * Return the degree of a given vertex
	 **/
	public int vertexDegree(int v) {
		int result=1;
		int eFirst=this.incidentEdge[v]; // edge incident to the vertex
		int pEdge; // edge turning around vertex v
		
		if(this.source[eFirst]==v) // if the edge is outgoing
			pEdge=this.rightBack[eFirst];
		else
			pEdge=this.leftFront[eFirst];
		
		while(pEdge!=eFirst) {
			if(this.source[pEdge]==v){ // if the edge is outgoing
				pEdge=this.rightBack[pEdge];
			}
			else
				pEdge=this.leftFront[pEdge];
			result++;
		}
		return result;	
	}
	
	/** 
	 * Return the degree of a given vertex
	 * This version is slightly slower (because of function calls)
	 **/
	public int vertexDegreeSlow(int v) {
		int result=1;
		int eFirst=this.getEdge(v); // edge incident to the vertex
		int pEdge; // edge turning around vertex v
		
		if(this.isOutgoing(eFirst, v)==true)
			pEdge=this.getRightBack(eFirst);
		else
			pEdge=this.getLeftFront(eFirst);
		
		while(pEdge!=eFirst) {
			//System.out.print("pEdge "+pEdge+": "+this.getSource(pEdge)+"-"+this.getTarget(pEdge));
			if(this.isOutgoing(pEdge, v)==true) {
				//System.out.print(" outgoing -> "+this.getRightBack(pEdge));
				pEdge=this.getRightBack(pEdge);
			}
			else
				pEdge=this.getLeftFront(pEdge);
			//System.out.println();
			result++;
		}
		return result;	
	}
		
	/**
	 * Return the normal to a vertex in the mesh (simple float precision)
	 */
	public float[] getNormalFloat(int v) {
		//System.out.println("Vertex normal winged edge"); // debug
		float[] result=new float[3]; // average of cross product vectors
		float[] faceNormal; // cross product vector
		int degree=1; // vertex degree
		int u, w; // indices of vertices u and v
		float[] pV, pU, pW; // coordinates of vertices v, u and w

		int edge=this.getEdge(v); 	// edge incident to vertex v
		int pEdge;					// edge turning around vertex v
		if(this.getSource(edge)==v) { // the edge is outgoing from v
			pEdge=this.getRightBack(edge);
			u=this.getTarget(edge);
			v=this.getSource(edge);
		}
		else { // the edge is ingoing (having v as target vertex)
			pEdge=this.getLeftFront(edge);
			v=this.getTarget(edge);
			u=this.getSource(edge);
		}
		if(this.getSource(pEdge)==v) // pEdge is outgoing from v
			w=this.getTarget(pEdge);
		else // pEdge is ingoing (having v as target vertex)
			w=this.getSource(pEdge);

		pV=this.coordF[v]; // coordinates of vertex v
		pU=this.coordF[u]; // coordinates of vertex u
		pW=this.coordF[w]; // coordinates of vertex w
		
		//System.out.println(vectorToString(pV)+" - "+vectorToString(pU)+" - "+vectorToString(pW)); // debug

		// computing vectors V-U and W-U
		result=NavigationInterface.computeFaceNormal(pU, pV, pW);
		pU=pW;
		
		while(pEdge!=edge) {
			//System.out.println("pEdge "+pEdge+" - edge "+edge);
			if(this.getSource(pEdge)==v) {
				pEdge=this.getRightBack(pEdge);
			}
			else {
				pEdge=this.getLeftFront(pEdge);
			}
			if(this.getSource(pEdge)==v) // pEdge is outgoing from v
				w=this.getTarget(pEdge);
			else // pEdge is ingoing (having v as target vertex)
				w=this.getSource(pEdge);
			
			pW=this.coordF[w]; // updating coordinates of vertex w	
			
			//System.out.println(vectorToString(pV)+" - "+vectorToString(pU)+" - "+vectorToString(pW)); // debug
			
			faceNormal=NavigationInterface.computeFaceNormal(pU, pV, pW);

			result[0]=result[0]+faceNormal[0];
			result[1]=result[1]+faceNormal[1];
			result[2]=result[2]+faceNormal[2];

			pU=pW;
			degree++;
		}	
		result[0]=result[0]/degree;
		result[1]=result[1]/degree;
		result[2]=result[2]/degree;
		
		//System.out.println("degree: "+degree);
		return result;			
	}
	
	//--- methods for visualizing the data structure ---
	
	/**
	 * Return a String describing the data structure references
	 */
	public String toString() {
		String result="Array based Winged edge DS\n";
		
		for(int e=0;e<this.nEdges;e++) {
			result=result+"edge \t"+e+": ("+this.getSource(e)+","+this.getTarget(e)+") -\t";
			
			int lF=this.getLeftFront(e);
			result=result+lF+"["+this.getSource(lF)+","+this.getTarget(lF)+"]   ";

			int rF=this.getRightFront(e);
			result=result+rF+"["+this.getSource(rF)+","+this.getTarget(rF)+"] \t";

			int lB=this.getLeftFront(e);
			result=result+lB+"["+this.getSource(lB)+","+this.getTarget(lB)+"] ";

			int rB=this.getRightFront(e);
			result=result+rB+"["+this.getSource(rB)+","+this.getTarget(rB)+"]";

			result=result+"\n";
		}
		
		return result;
	}
	
	/**
	 * Return the name of the data structure
	 */
	public String name() {
		return "Winged edge";
	}

	  public String vectorToString(float[] u) {
		  return "("+u[0]+","+u[1]+","+u[2]+")";
	  }

}
