package arrays;

/**
 * @author Luca Castelli Aleardi (Ecole Polytechnique, 2016)
 * <p>
 * This class provides an implementation of a stack storing integers: the elements are stored in
 * a given column (chosen by the user) of an already existing array A: <p>
 * 
 * The array A is assumed to be "ideally" subdivided into columns
 */
public class ColumnArrayBasedStack implements StackInterface {
    private int top;
    private final int maxSize; // maximal number of elements to be stored into the stack
    
    private int[] A; // an array, which is ideally subdivided into columns
    private int nColumns; // total number of columns of the array A
    private int column; // the index of the column used to store the elements in the stack
    private int nLines; // number of lines, which is equal to size(A)/nColumns
    
	/**
	 * Initialize a stack using an existing T: the elements will be stored in the i-th column of the array
	 * 
	 * @param column  the index of the column where the values will be stored
	 * @param A  input array, ideally subdivided into columns
	 * @param nColumns  number of columns of the array A
	 * @param size  maximal number of elements to be stored into the stack
	 */
    public ColumnArrayBasedStack(int[] A, int column, int nColumns, int size) {
    	this.maxSize=size;
    	this.A=A;
    	this.top=-1;
    	this.column=column;
    	this.nColumns=nColumns;
    	
    	this.nLines=A.length/nColumns;
    	if(A.length%nColumns!=0)
    		throw new Error("Error: the number of columns is incorrect "+nColumns);
    	if(this.nLines<this.maxSize)
    		throw new Error("Error: the memory allocated in T is too small ");
    }
    
	/**
	 * Pushes a value onto the top of this stack.
	 */
    public void push(int value) {
        if(top==maxSize-1)
        	throw new Error("Stack error: the stack is full");
        
        this.top++;
        int index=this.nColumns*this.top+this.column; // index of the case in array A, where the value will be stored
        this.A[index]=value;
    	//System.out.println("push "+value+" - "+this.toString());
    }

	/**
	 * Removes the value at the top of this stack and returns that value as the value of this function.
	 */
    public int pop() {
        if(!isEmpty()) {
            int index=this.nColumns*this.top+this.column; // index of the case in array A, where the value will be stored
        	int result=this.A[index];
            this.top--;
        	//System.out.println("pop "+result+" - "+this.toString());
            return result;
        }
        else
        	throw new Error("Stack error (pop call): stack empty");
    }

	/**
	 * Returns the value at the top of this stack (with deletion).
	 */
    public int top() {
        if(!isEmpty()) {
            int index=this.nColumns*this.top+this.column; // index of the case in array A, where the value will be stored
        	return this.A[index];
        }
        else
        	throw new Error("Stack error (pop call): stack empty");
    }

	/**
	 * Returns the numbers of elements stored in the stack.
	 */
    public int size() {
    	return this.top+1;
    }

	/**
	 * Check whether the stack is empty
	 */
    public boolean isEmpty() {
    	if(this.top==-1)
    		return true;
    	return false;
    }

	/**
	 * Reset the array: all elements are removed (set to 0)
	 */
    public void reset() {
    	int n=this.maxSize;
    	for(int i=0;i<n;i++) {
    		int index=this.nColumns*i+this.column; // index of the i-th element stored in the stack
    		this.A[index]=0;
    	}
    	
    	this.top=-1;
    }

	/**
	 * Returns a string representing the elements stored in the stack
	 */
    public String toString() {
    	String result="stack: ";
    	if(this.isEmpty())
    		return result+"empty";
    	
    	int n=this.size();
    	for(int i=n-1;i>=0;i--) { // show the elements in reverse order
    		int index=this.nColumns*i+this.column; // index of the i-th element stored in the stack
    		result=result+this.A[index]+"["+index+"] ";
    	}
    	return result;
    }

}
