package arrays;

import compactTriangulations.test.MemoryBenchmark;

/**
 * @author Luca Castelli Aleardi (Ecole Polytechnique, 2016)
 * 
 * Implementation of an array of integer references allowing to store 2 service bits per reference
 * <p>
 * Every integer is encoded on 32 bits and stores two informations: <p> 
 * -) a reference on 30 bits <p>
 * -) 2 service bits (the two rightmost bits: less significant bits)
 */
public class TableWithTwoServiceBits {

	/** The array whose elements are 32 bits integers: they contain a reference (on 30 bits) and the two service bits*/
	public int[] elements;

	/**
	 * Initialize the array (static array with 'nElements')
	 */
	public TableWithTwoServiceBits(int nElements) {
		this.elements=new int[nElements];
	}

	/**
	 * Retrieve the integer reference stored at position 'index'
	 * 
	 * @param index  the index (position) of the reference in the array
	 * @return the integer reference stored at position 'index' (on 30 bits)
	 */
	public int elementAt(int index) {
		return (this.elements[index]>>2);
	}
	
	/**
	 * Set the value of the integer reference stored at position 'index' <p>
	 * <p>
	 * Warning: <p> 
	 * -) it works only if the element is not already set <p>
	 * -) the two service bits are deleted: they are set to '0'
	 * 
	 * @param index  the index (position) of the reference in the array
	 * @param element  the value of the reference (an integer on 30 bits)
	 */
	public void set(int index, int element) {
		this.elements[index]=(element<<2);
	}
	
	public int getFirstServiceBit(int index) {
		// remark: the mask '& 0x01', means 00000001
		  return elements[index] & 0x01; // first service bit is stored at last bit of the reference
	}

	public int getSecondServiceBit(int index) {
		  return ((int)(elements[index] >>> 1)) & 0x01;
	}
	
	/**
	 * Set the first service bit to '1'
	 */
	public void setFirstServiceBit(int index) {
		elements[index]=elements[index]+1;
	}
	
	/**
	 * Set the second service bit to '1'
	 */
	public void setSecondServiceBit(int index) {
	    elements[index] += 2;		
	}
	
	/** 
	 * Return a string representing the i-th element stored in the table.
	 * <p>
	 * The result contains an integer value, and the two service bits. 
	 * Service bits are printed from the right: the right-most bit represents the first service bits
	 **/
	public String elementToString(int i) {
		String result="(";
		result=result+this.elementAt(i)+",";
		result=result+this.getSecondServiceBit(i)+",";
		return result+this.getFirstServiceBit(i)+")";
	}
	
	// below are some auxiliary functions (for handling references and service bits stored in an integer value)
	
	/**
	 * Retrieve the integer reference stored in an integer value 'v'
	 * 
	 * @param v  the input value
	 * @return the reference stored in 'v' (on 30 bits)
	 */
	public static int extractReferenceFromInteger(int v) {
		return (v>>2);
	}

	/**
	 * Retrieve the two service bits stored in 'v'
	 * 
	 * @param v  the input value
	 * @return  an integer (on two bits) representing the two service bits stored in 'v'
	 */
	public static int extractTwoServiceBitsFromInteger(int v) {
		return v & (0x01+0x02); // the mask '(0x01+0x02)' means 00000011
	}

	/**
	 * Store a value 'ref' (on 30 bits) and two service bits into an integer (on 32 bits)
	 * 
	 * @param element  the value of the reference to be stored (on 30 bits)
	 * @param serviceBits an integer (it must be in 0..3) representing service bits: '2' means '10', '3' means '11', '1' means 01, '0' means '00'
	 * @return an integer storing: (ref, bit1, bit0)
	 */
	public static int storeReferenceAndServiceBitsIntoInteger(int ref, int serviceBits) {
		int result=ref<<2;
		return result+serviceBits;
	}
	
}
